package app.views.match.components;

import app.data.bean.CalculatedMatch;
import app.data.entity.Game;
import app.data.entity.Match;
import app.data.service.ChessComService;
import app.data.service.MatchService;
import app.gameimage.GameImageService;
import app.navigation.Navigation;
import app.utils.ChessComUtils;
import app.utils.EntityComponentUtils;
import app.utils.StringUtils;
import app.views.navigation.interfaces.ContentConfigurable;
import com.vaadin.flow.component.ClickEvent;
import com.vaadin.flow.component.ComponentEventListener;
import com.vaadin.flow.component.UI;
import com.vaadin.flow.component.button.Button;
import com.vaadin.flow.component.button.ButtonVariant;
import com.vaadin.flow.component.html.Div;
import com.vaadin.flow.component.html.Label;
import com.vaadin.flow.component.icon.Icon;
import com.vaadin.flow.component.icon.VaadinIcon;
import com.vaadin.flow.component.orderedlayout.FlexComponent;
import com.vaadin.flow.component.orderedlayout.FlexLayout;
import com.vaadin.flow.component.orderedlayout.HorizontalLayout;
import com.vaadin.flow.component.orderedlayout.VerticalLayout;
import com.vaadin.flow.component.progressbar.ProgressBar;
import com.vaadin.flow.component.textfield.TextField;
import com.vaadin.flow.shared.Registration;
import org.springframework.beans.factory.annotation.Autowired;

import java.util.ArrayList;
import java.util.List;
import java.util.NoSuchElementException;
import java.util.Optional;

public class MatchComponent extends Div implements ContentConfigurable {

    private final Navigation navigation;
    private final ChessComService chessComService;
    private final MatchService matchService;
    private final GameImageService gameImageService;

    private final VerticalLayout headerLayout = new VerticalLayout();
    private final HorizontalLayout headerPlayersLayout = new HorizontalLayout();
    private final Label headerResultLabel = new Label();

    private final FlexLayout gamesLayout = new FlexLayout();
    private final VerticalLayout noGamesLayout = new VerticalLayout();

    private final FlexLayout editLayoutWrapper = new FlexLayout();
    private final VerticalLayout editLayout = new VerticalLayout();
    private final ArrayList<TextField> editTextFields = new ArrayList<>();
    // TODO: autocorrect "/live/game" to "/game/live/"
    private final Button editSubmitButton = new Button("Submit", new Icon(VaadinIcon.CHECK));
    // TODO: disable when no 6 values in form, and make sure you don't freeze forever when there are wrong entries
    // TODO: use overlapping progress here
    private Registration editSubmitButtonRegistration;
    private final Button editCancelButton = new Button("Cancel", new Icon(VaadinIcon.CLOSE));
    private Registration editCancelButtonRegistration;
    private final Button chessComButton = new Button("Autofill with the latest 6 games between the players", new Icon(VaadinIcon.MAGIC));
    private Registration chessComButtonButtonRegistration;

    private Match match;
    private CalculatedMatch calculatedMatch;

    public MatchComponent(Navigation navigation,
                          @Autowired ChessComService chessComService,
                          @Autowired GameImageService gameImageService) {
        this.navigation = navigation;
        this.chessComService = chessComService;
        this.matchService = navigation.getMatchService();
        this.gameImageService = gameImageService;

        add(headerLayout);

        defineHeaderLayout();
        defineGamesLayout();
        defineNoGamesLayout();
        defineEditLayout();
    }

    private void defineNoGamesLayout() {
        noGamesLayout.add(createNoGamesLabel(), createAddGamesButton());
        noGamesLayout.setAlignItems(FlexComponent.Alignment.CENTER);
    }

    private Label createNoGamesLabel() {
        return new Label("No games for this match in the database.");
    }

    private Button createAddGamesButton() {
        Button button = new Button("Add games", new Icon(VaadinIcon.PLUS));
        button.addThemeVariants(ButtonVariant.LUMO_PRIMARY);
        button.addClickListener(event -> {
            navigation.setEditFlag(true);
            remove(noGamesLayout);
        });
        return button;
    }

    private void defineHeaderLayout() {
        headerLayout.add(headerPlayersLayout, headerResultLabel);

        headerPlayersLayout.addClassName("match-header-players");

        headerResultLabel.addClassName("match-header-result");
        headerResultLabel.addClassName("bold-header");
    }

    private void defineGamesLayout() {
        gamesLayout.setFlexWrap(FlexLayout.FlexWrap.WRAP);
        gamesLayout.setJustifyContentMode(FlexComponent.JustifyContentMode.CENTER);
    }

    private void defineEditLayout() {
        defineEditLayoutWrapper();
        defineEditLayoutItself();
        defineEditLayoutTextFields();
        defineEditSubmitButton();
        addButtonsToEditLayout();
    }

    private void defineEditLayoutWrapper() {
        editLayoutWrapper.add(editLayout);
        editLayoutWrapper.setJustifyContentMode(FlexComponent.JustifyContentMode.CENTER);
    }

    private void defineEditLayoutItself() {
        editLayout.setAlignItems(FlexComponent.Alignment.CENTER);
        editLayout.setWidth("");

        editLayout.addClassName("wrapper");

        editLayout.add(chessComButton);
    }

    private void defineEditLayoutTextFields() {
        for (int i = 0; i < 6; i++) {
            TextField textField = new TextField();
            textField.setWidth("23em");
            HorizontalLayout horizontalLayout = new HorizontalLayout(new Label(String.format("Game %d:", i + 1)), textField);
            horizontalLayout.setAlignItems(FlexComponent.Alignment.CENTER);
            horizontalLayout.setJustifyContentMode(FlexComponent.JustifyContentMode.CENTER);
            editLayout.add(horizontalLayout);
            editTextFields.add(textField);
        }
    }

    private void defineEditSubmitButton() {
        editSubmitButton.addThemeVariants(ButtonVariant.LUMO_PRIMARY);
    }
    private void addButtonsToEditLayout() {
        HorizontalLayout buttonLayout = new HorizontalLayout();
        buttonLayout.setAlignItems(FlexComponent.Alignment.CENTER);
        buttonLayout.setJustifyContentMode(FlexComponent.JustifyContentMode.CENTER);
        buttonLayout.add(editSubmitButton, editCancelButton);
        editLayout.add(buttonLayout);
    }

    @Override
    public void configureContent() {
        try {
            match = navigation.getSelectedMatch().orElseThrow();
            calculatedMatch = matchService.getCalculatedMatch(match);
            configureHeaderContent();
            configureMainContent();
        } catch (NoSuchElementException e) {
            gamesLayout.removeAll();
            add(navigation.getValidationLabel());
        }
    }

    private void configureHeaderContent() {
        configureHeaderPlayersLayout();
        configureHeaderResultLabel();
    }

    private void configureHeaderPlayersLayout() {
        HorizontalLayout player1 = EntityComponentUtils.getPlayerLabel(calculatedMatch.getPlayer1());
        HorizontalLayout player2 = EntityComponentUtils.getPlayerLabel(calculatedMatch.getPlayer2());
        Label vs = new Label("vs.");

        headerPlayersLayout.removeAll();
        headerPlayersLayout.add(player1, vs, player2);
    }

    private void configureHeaderResultLabel() {
        if (match.getGames().isEmpty() || navigation.editFlag()) {
            headerLayout.remove(headerResultLabel);
        } else {
            headerLayout.add(headerResultLabel);
            headerResultLabel.setText(String.format("%s",
                    StringUtils.getResultString("-", calculatedMatch.getScore1(), calculatedMatch.getScore2())));
        }
    }

    private void configureMainContent() {
        if (navigation.editFlag()) {
            remove(gamesLayout);
            add(editLayoutWrapper);
            configureEditContent();
            return;
        }
        remove(editLayoutWrapper);
        configureGamesContent();
    }

    private void configureEditContent() {
        List<Game> games = new ArrayList<>(match.getGames());

        clearTextFields();
        fillTextFieldsWithURLs(games);

        configureChessComButton();
        configureEditSubmitButton();
        configureEditCancelButton();
    }

    private void clearTextFields() {
        editTextFields.forEach(textField -> textField.setValue(""));
    }

    private void fillTextFieldsWithURLs(List<Game> games) {
        for (int i = 0; i < Math.min(games.size(), 6); i++)
            editTextFields.get(i).setValue(ChessComUtils.getGameURL(games.get(i)));
    }

    private void configureChessComButton() {
        if (chessComButtonButtonRegistration != null) chessComButtonButtonRegistration.remove();
        chessComButtonButtonRegistration = chessComButton.addClickListener(createChessComButtonClickListener());
    }

    private ComponentEventListener<ClickEvent<Button>> createChessComButtonClickListener() {
        return event -> {
            List<Game> gamesBetweenPlayers = chessComService.getLatestGamesBetweenPlayers(match, 6, 2);
            fillTextFieldsWithURLs(gamesBetweenPlayers);
        };
    }

    private void configureEditSubmitButton() {
        if (editSubmitButtonRegistration != null) editSubmitButtonRegistration.remove();
        editSubmitButtonRegistration = editSubmitButton.addClickListener(createEditSubmitButtonClickListener());
    }

    private ComponentEventListener<ClickEvent<Button>> createEditSubmitButtonClickListener() {
        return event -> {
            match.getGames().clear();
            addGamesToMatchFromFieldValues();
            matchService.update(match);
            match.getGames().forEach(gameImageService::createImageIfNotPresent);

            navigation.setEditFlag(false);
        };
    }

    private void addGamesToMatchFromFieldValues() {
        for (TextField textField : editTextFields) {
            Optional<Game> game = chessComService.getGame(textField.getValue(), match); // TODO: handle this when Optional is empty!
            game.ifPresent(value -> match.getGames().add(value));
        }
    }

    private void configureEditCancelButton() {
        if (editCancelButtonRegistration != null) editCancelButtonRegistration.remove();
        editCancelButtonRegistration = editCancelButton.addClickListener(createEditCancelButtonListener());
    }

    private ComponentEventListener<ClickEvent<Button>> createEditCancelButtonListener() {
        return event -> UI.getCurrent().navigate(String.format("matchday/%s", navigation.getWildcardParam().replace("edit/", "")));
    }

    private void configureGamesContent() {
        if (match.getGames().isEmpty()) {
            remove(gamesLayout);
            add(noGamesLayout);
            return;
        }

        remove(noGamesLayout);
        add(gamesLayout);

        gamesLayout.removeAll();
        for (Game game : match.getGames()) {
            gamesLayout.add(new GameComponent(game, gameImageService));
        }
    }
}
